<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Malin Lorch <mlorch@liquidweb.com>
 * @date    Aug 12, 2020
 */

/**
 * @package Interworx
 * @subpackage Plugin_Acronis
 */
class Model_Backups_List extends Model_Acronis
{
    const BACKUP_ID = 'backupId';
    const CREATED = 'created';
    const DATE = 'date';
    const ARCHIVE_ID = 'archiveId';
    const VAULT_ID  = 'vaultId';
    const MACHINE_ID = 'machineId';
    const LEGACY_ID = 'legacyId';
    const TENANT_AGENT_ID = 'tenantAgentId';

    private $_instanceId;

    /**
     * @param string $instanceId
     */
    public function __construct($instanceId)
    {
        $this->_instanceId = $instanceId;
    }

    /**
     * Get backups sorted by date - newest first.
     *
     * @return array
     */
    protected function _getBackupsSortedByDate()
    {
        $backups = $this->_getApiDetailsFromCache();

        usort($backups, function ($backup1, $backup2) {
            $time1 = strtotime($backup1[Model_ApiResponse_Backup::API_CREATED]);
            $time2 = strtotime($backup2[Model_ApiResponse_Backup::API_CREATED]);

            return $time2 - $time1;
        });

        return array_values($backups);
    }

    /**
     * @return array
     */
    private function _getApiDetailsFromCache()
    {
        $backups = $this->_getBackupsFromCache();

        $apiDetails = array_map(function ($backup) {
            return $backup[Model_Backups_Backup::CACHE_API_DETAILS];
        }, $backups);

        return array_values($apiDetails);
    }

    /**
     * @return array
     */
    private function _getBackupsFromCache()
    {
        $Cache = $this->_getBackupsCache();

        if ($Cache->empty()) {
            $this->_cacheBackupsFromApi();
        }

        return $Cache->getAll();
    }

    /**
     * Cache backup information if backup id does
     * not already exist in cache.
     */
    private function _cacheBackupsFromApi()
    {
        $Backups = $this->_getBackupsFromApi();

        foreach ($Backups as $Backup) {
            $Backup->addToCache();
        }
    }



    /**
     * Cache any new backups.
     */
    protected function _updateBackups()
    {
        if ($this->_unmappedPrescriptDataExists()) {
            $this->_deletePrescriptsIfExpired();
            $this->_cacheBackupsFromApi();
            $this->_mapPrescriptsToBackups();
        }
    }

    /**
     * Check if unmapped prescript data exists.
     *
     * @return boolean
     */
    private function _unmappedPrescriptDataExists()
    {
        $Cache = $this->_getPrescriptCache();
        return !$Cache->empty();
    }

    /**
     * Delete any unmapped prescript data that is considered expired
     */
    private function _deletePrescriptsIfExpired()
    {
        $BackupsCache = $this->_getBackupsCache();
        $PrescriptCache = $this->_getPrescriptCache();

        foreach ($PrescriptCache->getAll() as $prescript) {
            $prescriptTimestamp = $prescript[$PrescriptCache::TIMESTAMP];

            if ($BackupsCache->expired($prescriptTimestamp)) {
                $PrescriptCache->delete($prescriptTimestamp);
            }
        }
    }

    /**
     * @return Model_Cache_Prescript
     */
    protected function _getPrescriptCache()
    {
        return new Model_Cache_Prescript();
    }


    /**
     * Map backups in api response to backup details cached in pre-backup script.
     */
    private function _mapPrescriptsToBackups()
    {
        $Cache = $this->_getPrescriptCache();
        $prescripts = $Cache->getAll();

        foreach ($prescripts as $prescript) {
            $this->_addPrescriptDataToBackupCache($prescript);
        }
    }

    /**
     * Add prescript data to cache.
     *
     * @param array $prescript
     */
    private function _addPrescriptDataToBackupCache(array $prescript)
    {
        try {
            $prescriptTimestamp = $prescript[Model_Cache_Prescript::TIMESTAMP];
            $Backup = $this->_getMatchingBackup($prescriptTimestamp);
            $Backup->addPrescriptDataToCache($prescript);
            $this->_deletePrescript($prescriptTimestamp);
        } catch (Exception_Acronis $e) {
            $e->logError();
        }
    }

    /**
     * @param mixed $timestamp
     */
    private function _deletePrescript($timestamp)
    {
        $Cache = $this->_getPrescriptCache();
        $Cache->delete($timestamp);
    }

    /**
     * Add prescript data to backup created within an acceptable time frame.
     *
     * @param string $timestampPrescript
     * @return Model_Backups_Backup
     */
    private function _getMatchingBackup($timestampPrescript)
    {
        $Backups = $this->_getBackupsWithoutPrescriptInfo();

        // allow timestamp of backup to be up to a minute earlier
        // in case server time is different from acronis time
        $minSeconds = -60;
        $maxSeconds = 120;
        $acceptableTimeFrame = range($minSeconds, $maxSeconds);

        foreach ($Backups as $Backup) {
            $timeDiff = $Backup->getTimeStamp() - $timestampPrescript;

            if (in_array($timeDiff, $acceptableTimeFrame)) {
                return $Backup;
            }
        }

        $created = Model_Acronis::getDateString($timestampPrescript);
        throw new Exception_Acronis_Prescript_NoBackup($created);
    }

    /**
     * Get all backups with prescript info missing.
     *
     * @return Model_Backups_Backup[]
     */
    private function _getBackupsWithoutPrescriptInfo()
    {
        $Cache = $this->_getBackupsCache();
        $backups = $Cache->getBackupsWithoutPrescriptInfo();

        return array_map(function ($backup) {
            $apiInfo = $backup[Model_Cache_Backup::API_DETAILS];
            return new Model_Backups_Backup($apiInfo);
        }, $backups);
    }

    /**
     * Get backups from api.
     *
     * @return Model_Backups_Backup[]
     */
    private function _getBackupsFromApi()
    {
        $backupsApi = $this->_getApi()->getBackups($this->_instanceId);

        $Backups = array();

        foreach ($backupsApi as $backup) {
            $BackupApi = new Model_ApiResponse_Backup($backup);
            $Backups[] = new Model_Backups_Backup($BackupApi->toArray());
        }

        return $Backups;
    }

    /**
     * @param array $args
     * @return Model_Backups_Api
     */
    protected function _getApi(array $args = null)
    {
        return parent::_getApi();
    }
}
