<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Malin Lorch <mlorch@liquidweb.com>
 * @date    Aug 14, 2020
 */

/**
 * @package Interworx
 * @subpackage Plugin_Acronis
 */
abstract class Model_Recovery_Files_Databases extends Model_Recovery_Files
{
    const UNMAPPED = 'UNMAPPED';
    const DUMPS_DESTINATION_PATH = self::NW_DOWNLOAD_DIR_PATH . '/db-dumps';
    const DUMPS_TEMP_PATH = '/home/.acronis_tmp/pre-acronis-mysql';

    const DUMP_INFO_USER = 'unixUser';
    const DUMP_INFO_DBS = 'dbNames';

    private $_restoreId;
    private $_dbCache;

    /**
     * @param string $backupId
     * @param string $restoreId
     */
    public function __construct($backupId, $restoreId = null)
    {
        $this->_restoreId = $restoreId;
        parent::__construct($backupId);
    }

    /**
     * Save dumps info from api in cache.
     */
    protected function _cacheDumpsFromApi()
    {
        IW::getLog()->warning('Acronis: Databases could not be retrieved from prescript - attempting to get db names from acronis api.');

        $dbInfo = $this->_getDumpsInfoFromApi();
        $this->_getCache()->cacheDumps($this->getBackupId(), $dbInfo);
    }

    /**
     * Get dumps info.
     *
     * @return array
     */
    private function _getDumpsInfoFromApi()
    {
        try {
            $dumpsInfo = $this->_getApiDumpsInfo();
            return $dumpsInfo;
        } catch (Exception_Acronis_InvalidPath $e) {
            if (isset($this->_restoreId)) {
                $this->_setFailed('Database recovery not enabled');
            }

            throw new Exception_Acronis_DbRecoveryNotEnabled();
        }
    }

    /**
     * get dumps info from api.
     * Will throw exception if dumps do not exist.
     *
     * @return array
     */
    private function _getApiDumpsInfo()
    {
        $content = $this->_getDirContentByPath(self::DUMPS_TEMP_PATH);
        $result = array_map(function ($dir) {
            $name = $dir['name'];

            if (strpos($name, '-dumps') !== false) {
                $path = self::DUMPS_TEMP_PATH . "/$name";
                return $this->_getUserDbs($path, $name);
            }
        }, $content);

        return array_values(array_filter($result));
    }

    /**
     * Get db info, user & db names.
     *
     * @param string $path
     * @param string $dirName
     */
    private function _getUserDbs($path, $dirName)
    {
        if (strpos($dirName, '-dumps') === false) {
            throw new Exception('not dbs folder: ' . $dirName);
        }

        $dbUser = str_replace('-dumps', '', $dirName);
        $dbNames = $this->_getNamesFromDirContentByPath($path);

        return array(
            self::DUMP_INFO_USER => $dbUser,
            self::DUMP_INFO_DBS => $dbNames
        );
    }

    /**
     * Excludes mysql db.
     *
     * @param string $path
     */
    private function _getNamesFromDirContentByPath($path)
    {
        $content = $this->_getDirContentByPath($path);

        $result = array_map(function ($dir) {
            if ($dir['name'] !== 'mysql') {
                return $dir['name'];
            }
        }, $content);

        return array_values(array_filter($result));
    }

    /**
     * Run command as root and get output as array.
     *
     * @param string $cmd
     * @return array
     */
    protected function _getCmdOutput($cmd)
    {
        $Command = new Command($cmd);
        $Command->runAsRoot();
        $Command->allowAsterisk();
        $Command->allowAmpersand();
        $Command->allowRedirectOut();
        $Command->allowCurlyBoth();
        $Command->exec();
        return $Command->getOutput();
    }

    /**
     * Restore user dumps.
     *
     * @param string $unixUser
     */
    protected function _restoreUserDumps($unixUser)
    {
        IW::getLog()->info('Acronis: restoring db dumps for ' . $unixUser);

        $legacyId = $this->getLegacyId();
        $tenantAgentId = $this->getAgentTenantId();
        $path = $this->_getUserDumpsPathInBackups($unixUser);
        $Dir = $this->_getDirByPath($path);

        $ids = array($Dir->getId());
        $destination = $this->_getTempDestinationPath($unixUser);

        $this->_createRestoreAltLocationDir($destination);
        $Api = $this->_getRecoveryApi();
        $Api->recoverDirectoriesToAltLocation($legacyId, $tenantAgentId, $ids, $destination);
        $this->_cachePlanId();
    }

    /**
     * Get the temporary destination path for db restores.
     *
     * @param string $unixUser
     * @return string
     */
    protected function _getTempDestinationPath($unixUser)
    {
        return self::DUMPS_DESTINATION_PATH . "/$unixUser";
    }

    /**
     * Get user dumps path.
     *
     * @param string $unixUser
     */
    protected function _getUserDumpsPathInBackups($unixUser)
    {
        return self::DUMPS_TEMP_PATH . "/$unixUser-dumps";
    }

    /**
     * Await restoration. Will wait max 3 minutes for restoration to complete.
     *
     * @param string $unixUser
     */
    protected function _awaitUserDumpsRestoration($unixUser)
    {
        IW::getLog()->info('Acronis: awaiting dumps restoration for user ' . $unixUser);

        $count = 1;
        while (!$this->_userDumpsRestored($unixUser) && $count < 60) {
            $count++;
            sleep(60);
        }

        if (!$this->_userDumpsRestored($unixUser)) {
            // cancel process if file restoration
            // takes longer than 1 hr. Files may still
            // get restored at a later time, but InterWorx will
            // stop looking for it & complete the process.
            IW::getLog()->error('Acronis: acronis api took too long to restore files. Db restoration interrupted for user ' . $unixUser);
            $this->_setFailed('acronis failed to restore file.');
            throw new Exception('Dumps not restored to tmp location!');
        }
    }

    /**
     * Check if user dumps are restored.
     *
     * Empty directory is always created per user in pre script.
     * If no databases exists for the user, the folder will still
     * exist but be empty.
     *
     * @param string $unixUser
     */
    protected function _userDumpsRestored($unixUser)
    {
        $path = $this->_getRestoredDumpsPath($unixUser);
        return Model_AcronisFileSys::isDir($path);
    }

    /**
     * unzip xz archives to specified destination.
     *
     * @param string $archive
     * @param string $destination
     */
    protected function _unxz($archive, $destination)
    {
        $cmd = "unxz -c $archive > $destination && chown iworx:iworx $destination";
        IW::getLog()->info($cmd);

        $Command = new Command($cmd);
        $Command->runAsRoot();
        $Command->allowRedirectOut();
        $Command->allowAmpersand();
        $Command->allowColon();
        $Command->exec();
    }

    /**
     * unzip archives to specified destination.
     *
     * @param string $archive
     * @param string $destination
     */
    protected function _gunzip($archive, $destination)
    {
        $cmd = "gunzip -c $archive > $destination && chown iworx:iworx $destination";
        $Command = new Command($cmd);
        $Command->runAsRoot();
        $Command->allowRedirectOut();
        $Command->allowAmpersand();
        $Command->allowColon();
        $Command->exec();
    }

    /**
     * Uncompress archive.
     *
     * @param string $archive
     * @param string $destination
     */
    protected function _uncompress($archive, $destination)
    {
        $ext = pathinfo($archive, PATHINFO_EXTENSION);

        if ($ext == 'gz') {
            $this->_gunzip($archive, $destination);
        }
        if ($ext == 'xz') {
            $this->_unxz($archive, $destination);
        }
    }

    /**
     * Get restored dumps path - the directory that will be
     * created by acronis after successful restore.
     *
     * @param string $unixUser
     */
    protected function _getRestoredDumpsPath($unixUser)
    {
        $dumpsPath = $this->_getTempDestinationPath($unixUser) . "/$unixUser-dumps";
        return $dumpsPath;
    }

    /**
     * Get array of archives containing sql file for tables.
     *
     * @param string $parentFolder
     * @return array
     */
    protected function _getTableArchives($parentFolder)
    {
        return $this->_getCmdOutput("ls $parentFolder/*.{xz,gz} 2> /dev/null");
    }

    /**
     * get sql files in specified location.
     *
     * @param string $sqlPath
     * @param string $dbName
     * @return array
     */
    protected function _getSqlFiles($sqlPath, $dbName)
    {
        $cmd = "ls $sqlPath/$dbName*.sql 2> /dev/null";
        return $this->_getCmdOutput($cmd);
    }


    /**
     * cleanup dumps.
     *
     * @param string $unixUser
     */
    protected function _cleanupDumps($unixUser)
    {
        $dumpsPath = $this->_getTempDestinationPath($unixUser);
        $Command = new Command("rm -rf $dumpsPath");
        $Command->runAsRoot();
        $Command->exec();
    }

    /**
     *
     * @param string $user
     * @param array  $dbNames
     * @param string $destination
     */
    protected function _restoreUserDbsToAltLocation($user, array $dbNames, $destination)
    {
        $tenantAgentId = $this->getAgentTenantId();
        $legacyId = $this->getLegacyId();
        $this->_createRestoreAltLocationDir($destination);

        $dirIds = $this->_getDbIds($user, $dbNames);
        $this->_getRecoveryApi()->recoverDirectoriesToAltLocation($legacyId, $tenantAgentId, $dirIds, $destination);
    }

    /**
     * @param string $user
     * @param array  $databaseNames
     */
    private function _getDbIds($user, array $databaseNames)
    {
        $userDumpsPath = $this->_getUserDumpsPathInBackups($user);
        $dbDirs = $this->_getDirContentByPath($userDumpsPath);

        $ids = array();
        foreach ($dbDirs as $dir) {
            if (in_array($dir['name'], $databaseNames)) {
                $ids[] = $dir['id'];
            }
        }

        return $ids;
    }

    /**
     * @return Model_Cache_Restore_Databases
     */
    protected function _getDbCache()
    {
        if (!isset($this->_dbCache)) {
            $this->_dbCache = new Model_Cache_Restore_Databases();
        }

        return $this->_dbCache;
    }

    /**
     * @param array  $dbs
     * @param string $unixuser
     */
    protected function _addToRestoreCache(array $dbs, $unixuser)
    {
        if (!isset($this->_restoreId)) {
            throw new Exception('Acronis: Db restore: restore id missing');
        }

        $Cache = $this->_getDbCache();
        $cacheInfo = $this->_getCacheFormat($dbs, $unixuser);
        $Cache->add($this->_restoreId, $cacheInfo);
    }

    /**
     * @param array  $dbs
     * @param string $unixuser
     * @return array
     */
    private function _getCacheFormat(array $dbs, $unixuser)
    {
        return array(
            Model_Cache_Abstract::ID => $this->_restoreId,
            Model_Cache_Abstract::TIME => Model_Acronis::getDateString(time()),
            Model_Cache_Abstract::DBS => $dbs,
            Model_Cache_Abstract::UNIXUSER => $unixuser,
            Model_Cache_Restore::STATE => Model_Cache_Restore::STATE_STARTED
        );
    }

    /**
     * update state - in progress (db dumps are restored and we have started the restoration)
     */
    protected function _setInProgress()
    {
        $this->_getDbCache()->setInProgress($this->_restoreId);
    }

    /**
     * update state - completed
     */
    protected function _setComplete()
    {
        $this->_getDbCache()->setComplete($this->_restoreId);
    }

    /**
     * Update state - failed & add error message
     *
     * @param string $msg
     */
    protected function _setFailed($msg)
    {
        $this->_getDbCache()->setFailed($this->_restoreId);
        $this->_getDbCache()->setErrorMsg($this->_restoreId, $msg);
    }

    /**
     * Getting plan id from acronis api. Will retry in 10 sec if not updated at the time of exec.
     */
    protected function _cachePlanId()
    {
        $planId = $this->getPlanId();
        if ($planId === null) {
            sleep(10);
            $planId = $this->getPlanId();
        }

        $this->_getDbCache()->setPlanId($this->_restoreId, $this->getPlanId());
    }

    /**
     * @return string
     */
    public function getRestoreId()
    {
        return $this->_restoreId;
    }

    /**
     * @return Model_Cache_Backup
     */
    protected function _getBackupCache()
    {
        return new Model_Cache_Backup();
    }

    /**
     * @param array $dbs
     * @return array
     */
    protected function _getDbNames(array $dbs)
    {
        return array_map(function ($db) {
            return $db['name'];
        }, $dbs);
    }
}
