<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Malin Lorch <mlorch@liquidweb.com>
 * @date    Dec 04, 2020
 */
require_once(__DIR__ . '/Acronis.php');

/**
 * Test Suit Model_Server_Server
 *
 * @package Interworx
 * @subpackage Plugin_Acronis
 */
class Model_Server_ServerTest extends AcronisTest
{
    const CACHED_PLANID = "12345-CACHED";
    const SCRIPTS_OK_PLANID = "01A4CF6F-6628-44BF-B536-2D2F6E4D48FC";
    const INVALID_PLANID = "BJAJNKJDNSJDNJJNFJNFASJ";
    const INVALID_PLANID_2 = "BJAJNKJDNSJDNJ";

    private $_cache;

    /**
     * getPlanId should return cached plan id if exists in api.
     */
    public function testGetPlanIdShouldReturnCached()
    {
        $SUT = $this->_getSUT();

        // prep cache to return
        // cached plan id
        $this->getCacheMock()
            ->expects($this->any())
            ->method('getPlanId')
            ->willReturn(self::CACHED_PLANID);

        $actual = $SUT->getPlanId();
        $expected = self::CACHED_PLANID;

        $this->assertSame($actual, $expected);
    }


    /**
     * getPlanId should not return valid plan if cached does not exist in api.
     */
    public function testGetPlanIdShouldReturnValidPlan()
    {
        $expectedValue = self::SCRIPTS_OK_PLANID;
        $SUT = $this->_getSUT();

        // configure cache mock
        $nonexistingPlanId = 'somerandomnonexistingid';
        $CacheMock = $this->getCacheMock();
        $this->_configureCacheMockGetPlanId($CacheMock, $nonexistingPlanId);

        $this->_expectMockToCacheId($CacheMock, $expectedValue);
        $actual = $SUT->getPlanId();
        $this->assertNotSame($actual, $nonexistingPlanId);
        $this->assertSame($actual, $expectedValue);
    }


    /**
     * getPlanId should return random invalid plan if
     * no cache && no valid plans.
     */
    public function testGetPlanIdShouldReturnInvalidPlan()
    {
        $apiShouldReturnValidPlans = false;
        $SUT = $this->_getSUT($apiShouldReturnValidPlans);


        $CacheMock = $this->getCacheMock();
        $expectedValues = array(self::INVALID_PLANID, self::INVALID_PLANID_2);

        $this->_expectMockToCacheId($CacheMock, $expectedValues);
        $actual = $SUT->getPlanId();
        $this->assertContains($actual, $expectedValues);
    }

    /**
     * @param PHPUnit\Framework\MockObject\MockObject $CacheMock
     * @param string                                  $planId
     */
    private function _configureCacheMockGetPlanId(PHPUnit\Framework\MockObject\MockObject $CacheMock, $planId)
    {
        // mock cache to return plan id that
        // does not exist in json from api
        $CacheMock
            ->expects($this->once())
            ->method('getPlanId')
            ->willReturn($planId);
    }

    /**
     * Make sure cachePlanId is being called with the right value.
     *
     * @param PHPUnit\Framework\MockObject\MockObject $CacheMock
     * @param string|array                            $id
     */
    private function _expectMockToCacheId(PHPUnit\Framework\MockObject\MockObject $CacheMock, $id)
    {
        // if array, assume that we expect random plan id
        // as no cache && no valid plans were returned by api
        $expected = is_array($id) ? $this->_cbConfirmArgIsInvalidPlanId() : $id;

        $CacheMock
            ->expects($this->once())
            ->method('cachePlanId')
            ->with($expected);
    }

    /**
     * expects any of the invalid plan ids
     * if no cached && no valid plans in array from api
     */
    private function _cbConfirmArgIsInvalidPlanId()
    {
        return $this->callback(function ($arg) {
            $invalidPlanIds = array(self::INVALID_PLANID, self::INVALID_PLANID_2);
            return in_array($arg, $invalidPlanIds, true);
        });
    }

    /**
     * getApplicablePlanId should return cached plan when
     * it exists in api.
     */
    public function testGetApplicablePlanIdShouldReturnCache()
    {
        $SUT = $this->_getSUT();
        $expected = self::CACHED_PLANID;

        // prep cache to return
        // cached plan id
        $this->getCacheMock()
            ->expects($this->any())
            ->method('getPlanId')
            ->willReturn($expected);

        $actual = $SUT->getApplicablePlanId();
        $this->assertSame($actual, $expected);
    }


    /**
     * same as above but parsing api response without any valid plans.
     */
    public function testGetApplicablePlanIdShouldReturnCachedWithNoValidPlans()
    {
        $apiShouldReturnValidPlans = false;
        $SUT = $this->_getSUT($apiShouldReturnValidPlans);

        $expected = self::CACHED_PLANID;

        // prep cache to return
        // cached plan id
        $this->getCacheMock()
            ->expects($this->any())
            ->method('getPlanId')
            ->willReturn($expected);

        $actual = $SUT->getApplicablePlanId();
        $this->assertSame($actual, $expected);
    }

    /**
     * getApplicablePlanId should return id from valid plan
     * when it exists in api
     */
    public function testGetApplicablePlanIdShouldReturnValidPlan()
    {
        $SUT = $this->_getSUT();
        $expected = self::SCRIPTS_OK_PLANID;

        $this->getCacheMock()
            ->expects($this->any())
            ->method('getPlanId')
            ->willReturn('NONEEXISTINGPLANID');

        $actual = $SUT->getApplicablePlanId();
        $this->assertSame($actual, $expected);
    }

    /**
     * getApplicablePlan should throw exception if api does not include any applicable plan
     * && none is cached
     */
    public function testGetApplicablePlanIdShouldThrowException()
    {
        $apiShouldReturnValidPlans = false;
        $SUT = $this->_getSUT($apiShouldReturnValidPlans);

        $this->expectException(Exception::class);
        $SUT->getApplicablePlanId();
    }

    /**
     * Get system under test - partial mock of Model_Server_Server.
     *
     * @param boolean $validPlans
     */
    private function _getSUT($validPlans = true)
    {
        $methodsToMock = array(
            self::ACTION_GET_API,
            self::ACTION_GET_CACHE,
            'getInstanceId',
            'getResourceId',
            '_prePostScriptsOk',
            self::ACTION_GET_PLUGIN
        );

        $SUT = $this->getMockBuilder('Model_Server_Server')
            ->setMethods($methodsToMock)
            ->getMock();

        // configure sut
        $SUT->expects($this->any())
            ->method(self::ACTION_GET_CACHE)
            ->willReturn($this->getCacheMock());

        $SUT->expects($this->any())
            ->method(self::ACTION_GET_API)
            ->willReturn($this->getApiMock($validPlans));

        $SUT->expects($this->any())
            ->method('_prePostScriptsOk')
            ->willReturnMap(array(
                // return false or true
                // depending on argument
                [self::CACHED_PLANID, false],
                [self::INVALID_PLANID, false],
                [self::INVALID_PLANID_2, false],
                [self::SCRIPTS_OK_PLANID, true],
            ));


        return $SUT;
    }

    /**
     * get api mock.
     *
     * @param boolean $validPlans
     */
    private function getApiMock($validPlans = true)
    {
        $getPlansFile = $validPlans ? '/server/plans.json' : '/server/novalidplans.json';
        $getApplicablePlansFile = $validPlans ? '/server/applicablePlans.json' : '/server/applicablePlansNoValid.json';

        $jsonGetPlans = file_get_contents(__DIR__ . $getPlansFile);
        $jsonGetApplicablePlans = file_get_contents(__DIR__ . $getApplicablePlansFile);

        $Mock = $this->createMock('Model_Server_Api');

        $Mock->expects($this->any())
            ->method('getBackupPlans')
            ->willReturn(json_decode($jsonGetPlans, true));

        $Mock->expects($this->any())
            ->method('getApplicablePlans')
            ->willReturn(json_decode($jsonGetApplicablePlans, true));

        return $Mock;
    }

    /**
     * get cache mock.
     *
     * @return PHPUnit\Framework\MockObject\MockObject
     */
    private function getCacheMock()
    {
        if (!isset($this->_cache)) {
            $this->_cache = $this->_createCacheMock();
        }

        return $this->_cache;
    }
}
